@@ -1,29 +1,27 @@
-/*
- *  File:       libutil.cc
- *  Summary:    Functions that may be missing from some systems
- *  Written by: ?
- *
- *  Modified for Crawl Reference by $Author$ on $Date$
- */
+/**
+ * @file
+ * @brief Functions that may be missing from some systems
+**/
 
 #include "AppHdr.h"
-REVISION("$Rev$");
 
 #include "defines.h"
-#include "directn.h"
-#include "initfile.h"
 #include "itemname.h" // is_vowel()
 #include "libutil.h"
 #include "externs.h"
+#include "files.h"
 #include "macro.h"
-#include "stuff.h"
+#include "message.h"
+#include "unicode.h"
+#include "viewgeom.h"
+
 #include <sstream>
 #include <stdio.h>
 #include <ctype.h>
 #include <stdarg.h>
 #include <string.h>
 
-#if defined(WIN32CONSOLE) || defined(WIN32TILES)
+#ifdef TARGET_OS_WINDOWS
     #undef ARRAYSZ
     #include <windows.h>
     #undef max
@@ -33,19 +31,13 @@ REVISION("$Rev$");
     #endif
 #endif
 
-#ifdef REGEX_PCRE
-    // Statically link pcre on Windows
-    #if defined(WIN32CONSOLE) || defined(WIN32TILES) || defined(DOS)
-        #define PCRE_STATIC
-    #endif
-
-    #include <pcre.h>
+#ifdef UNIX
+    #include <signal.h>
 #endif
 
-#ifdef REGEX_POSIX
-    // Do we still need to include sys/types.h?
-    #include <sys/types.h>
-    #include <regex.h>
+#ifdef DGL_ENABLE_CORE_DUMP
+    #include <sys/time.h>
+    #include <sys/resource.h>
 #endif
 
 description_level_type description_type_by_name(const char *desc)
@@ -73,6 +65,10 @@ description_level_type description_type_by_name(const char *desc)
         return DESC_INVENTORY;
     else if (!strcmp("none", desc))
         return DESC_NONE;
+    else if (!strcmp("base", desc))
+        return DESC_BASENAME;
+    else if (!strcmp("qual", desc))
+        return DESC_QUALNAME;
 
     return DESC_PLAIN;
 }
@@ -116,12 +112,12 @@ bool shell_safe(const char *file)
     return (match < 0 || !file[match]);
 }
 
-void play_sound( const char *file )
+void play_sound(const char *file)
 {
 #if defined(WINMM_PLAY_SOUNDS)
     // Check whether file exists, is readable, etc.?
     if (file && *file)
-        sndPlaySound(file, SND_ASYNC | SND_NODEFAULT);
+        sndPlaySoundW(OUTW(file), SND_ASYNC | SND_NODEFAULT);
 
 #elif defined(SOUND_PLAY_COMMAND)
     char command[255];
@@ -130,7 +126,7 @@ void play_sound( const char *file )
         && shell_safe(file))
     {
         snprintf(command, sizeof command, SOUND_PLAY_COMMAND, file);
-        system(command);
+        system(OUTS(command));
     }
 #endif
 }
@@ -140,17 +136,76 @@ std::string strip_filename_unsafe_chars(const std::string &s)
     return replace_all_of(s, " .&`\"\'|;{}()[]<>*%$#@!~?", "");
 }
 
+std::string vmake_stringf(const char* s, va_list args)
+{
+    char buf1[8000];
+    va_list orig_args;
+    va_copy(orig_args, args);
+    size_t len = vsnprintf(buf1, sizeof buf1, s, orig_args);
+    va_end(orig_args);
+    if (len < sizeof buf1)
+        return (buf1);
+
+    char *buf2 = (char*)malloc(len + 1);
+    va_copy(orig_args, args);
+    vsnprintf(buf2, len + 1, s, orig_args);
+    va_end(orig_args);
+    std::string ret(buf2);
+    free(buf2);
+
+    return (ret);
+}
+
 std::string make_stringf(const char *s, ...)
 {
     va_list args;
     va_start(args, s);
-
-    char buf[400];
-    vsnprintf(buf, sizeof buf, s, args);
-
+    std::string ret = vmake_stringf(s, args);
     va_end(args);
+    return ret;
+}
 
-    return (buf);
+std::string &escape_path_spaces(std::string &s)
+{
+    std::string result;
+    result.clear();
+#ifdef UNIX
+    for (const char* ch = s.c_str(); *ch != '\0'; ++ch)
+    {
+        if (*ch == ' ')
+        {
+            result += '\\';
+        }
+        result += *ch;
+    }
+#elif defined(TARGET_OS_WINDOWS)
+    if (s.find(" ") != std::string::npos
+        && s.find("\"") == std::string::npos)
+    {
+        result = "\"" + s + "\"";
+    }
+    else
+    {
+        return s;
+    }
+#else
+    // Not implemented for this platform.  Assume that escaping isn't
+    // necessary.
+    return s;
+#endif
+    s = result;
+    return s;
+}
+
+bool key_is_escape(int key)
+{
+    switch (key)
+    {
+    CASE_ESCAPE
+        return (true);
+    default:
+        return (false);
+    }
 }
 
 std::string &uppercase(std::string &s)
@@ -194,6 +249,43 @@ int ends_with(const std::string &s, const char *suffixes[])
     return (0);
 }
 
+#ifdef UNIX
+extern "C" int stricmp(const char *str1, const char *str2)
+{
+    int ret = 0;
+
+    // No need to check for *str1.  If str1 ends, then tolower(*str1) will be
+    // 0, ret will be -1, and the loop will break.
+    while (!ret && *str2)
+    {
+        unsigned char c1 = tolower(*str1);
+        unsigned char c2 = tolower(*str2);
+
+        ret = c1 - c2;
+        str1++;
+        str2++;
+    }
+
+    if (ret < 0)
+        ret = -1;
+    else if (ret > 0)
+        ret = 1;
+
+    return (ret);
+}
+#endif
+
+bool strip_suffix(std::string &s, const std::string &suffix)
+{
+    if (ends_with(s, suffix))
+    {
+        s.erase(s.length() - suffix.length(), suffix.length());
+        trim_string(s);
+        return (true);
+    }
+    return false;
+}
+
 // Returns true if s contains tag 'tag', and strips out tag from s.
 bool strip_tag(std::string &s, const std::string &tag, bool skip_padding)
 {
@@ -236,6 +328,22 @@ bool strip_tag(std::string &s, const std::string &tag, bool skip_padding)
     return (false);
 }
 
+std::vector<std::string> strip_multiple_tag_prefix (std::string &s, const std::string &tagprefix)
+{
+    std::vector<std::string> results;
+
+    while (true)
+    {
+        std::string this_result = strip_tag_prefix(s, tagprefix);
+        if (this_result.empty())
+            break;
+
+        results.push_back(this_result);
+    }
+
+    return results;
+}
+
 std::string strip_tag_prefix(std::string &s, const std::string &tagprefix)
 {
     std::string::size_type pos = s.find(tagprefix);
@@ -345,10 +453,22 @@ std::string pluralise(const std::string &name,
         // Vortex; vortexes is legal, but the classic plural is cooler.
         return name.substr(0, name.length() - 2) + "ices";
     }
+    else if (ends_with(name, "mosquito"))
+    {
+        return name + "es";
+    }
     else if (ends_with(name, "cyclops"))
     {
         return name.substr(0, name.length() - 1) + "es";
     }
+    else if (name == "catoblepas")
+    {
+        return "catoblepae";
+    }
+    else if (ends_with(name, "s"))
+    {
+        return name;
+    }
     else if (ends_with(name, "y"))
     {
         if (name == "y")
@@ -383,10 +503,12 @@ std::string pluralise(const std::string &name,
         // mage -> magi
         return name.substr(0, name.length() - 1) + "i";
     }
-    else if (ends_with(name, "sheep") || ends_with(name, "manes")
-              || ends_with(name, "fish") || ends_with(name, "folk"))
+    else if (ends_with(name, "sheep") || ends_with(name, "fish")
+             || ends_with(name, "folk") || ends_with(name, "spawn")
+             || ends_with(name, "kenku") || ends_with(name, "shedu")
+             // "shedu" is male, "lammasu" is female of the same creature
+             || ends_with(name, "lammasu") || ends_with(name, "lamassu"))
     {
-        // Maybe we should generalise 'manes' to ends_with("es")?
         return name;
     }
     else if (ends_with(name, "ch") || ends_with(name, "sh")
@@ -396,9 +518,10 @@ std::string pluralise(const std::string &name,
         // ending with sh (except fish, which are caught in the previous check).
         return name + "es";
     }
-    else if (ends_with(name, "um"))
+    else if (ends_with(name, "simulacrum") || ends_with(name, "eidolon"))
     {
-        // simulacrum -> simulacra
+        // simulacrum -> simulacra (correct Latin pluralisation)
+        // also eidolon -> eidola (correct Greek pluralisation)
         return name.substr(0, name.length() - 2) + "a";
     }
     else if (ends_with(name, "efreet"))
@@ -408,10 +531,40 @@ std::string pluralise(const std::string &name,
     }
     else if (name == "foot")
         return "feet";
+    else if (name == "ophan" || name == "cherub" || name == "seraph")
+    {
+        // unlike "angel" which is fully assimilated and "cherub" which may be
+        // pluralized both ways, "ophan" always uses hebrew pluralization
+        return name + "im";
+    }
 
     return name + "s";
 }
 
+std::string apostrophise(const std::string &name)
+{
+    if (name.empty())
+        return (name);
+
+    if (name == "you" || name == "You")
+        return (name + "r");
+
+    if (name == "it" || name == "It")
+        return (name + "s");
+
+    const char lastc = name[name.length() - 1];
+    return (name + (lastc == 's' ? "'" : "'s"));
+}
+
+std::string apostrophise_fixup(const std::string &msg)
+{
+    if (msg.empty())
+        return (msg);
+
+    // XXX: This is rather hackish.
+    return (replace_all(msg, "s's", "s'"));
+}
+
 static std::string pow_in_words(int pow)
 {
     switch (pow)
@@ -485,12 +638,13 @@ std::string replace_all(std::string s,
                         const std::string &find,
                         const std::string &repl)
 {
+    ASSERT(!find.empty());
     std::string::size_type start = 0;
     std::string::size_type found;
 
     while ((found = s.find(find, start)) != std::string::npos)
     {
-        s.replace( found, find.length(), repl );
+        s.replace(found, find.length(), repl);
         start = found + repl.length();
     }
 
@@ -503,12 +657,13 @@ std::string replace_all_of(std::string s,
                            const std::string &tofind,
                            const std::string &replacement)
 {
+    ASSERT(!tofind.empty());
     std::string::size_type start = 0;
     std::string::size_type found;
 
     while ((found = s.find_first_of(tofind, start)) != std::string::npos)
     {
-        s.replace( found, 1, replacement );
+        s.replace(found, 1, replacement);
         start = found + replacement.length();
     }
 
@@ -517,6 +672,7 @@ std::string replace_all_of(std::string s,
 
 int count_occurrences(const std::string &text, const std::string &s)
 {
+    ASSERT(!s.empty());
     int nfound = 0;
     std::string::size_type pos = 0;
 
@@ -529,28 +685,28 @@ int count_occurrences(const std::string &text, const std::string &s)
     return (nfound);
 }
 
-std::string trimmed_string( std::string s )
+std::string trimmed_string(std::string s)
 {
     trim_string(s);
     return (s);
 }
 
 // also used with macros
-std::string &trim_string( std::string &str )
+std::string &trim_string(std::string &str)
 {
-    str.erase( 0, str.find_first_not_of( " \t\n\r" ) );
-    str.erase( str.find_last_not_of( " \t\n\r" ) + 1 );
+    str.erase(0, str.find_first_not_of(" \t\n\r"));
+    str.erase(str.find_last_not_of(" \t\n\r") + 1);
 
     return (str);
 }
 
-std::string &trim_string_right( std::string &str )
+std::string &trim_string_right(std::string &str)
 {
-    str.erase( str.find_last_not_of( " \t\n\r" ) + 1 );
+    str.erase(str.find_last_not_of(" \t\n\r") + 1);
     return (str);
 }
 
-static void add_segment( std::vector<std::string> &segs,
+static void add_segment(std::vector<std::string> &segs,
                          std::string s,
                          bool trim,
                          bool accept_empty)
@@ -562,7 +718,7 @@ static void add_segment( std::vector<std::string> &segs,
         segs.push_back(s);
 }
 
-std::vector<std::string> split_string( const std::string &sep,
+std::vector<std::string> split_string(const std::string &sep,
                                        std::string s,
                                        bool trim_segments,
                                        bool accept_empty_segments,
@@ -589,259 +745,389 @@ std::vector<std::string> split_string( const std::string &sep,
     return segments;
 }
 
-// The old school way of doing short delays via low level I/O sync.
-// Good for systems like old versions of Solaris that don't have usleep.
-#ifdef NEED_USLEEP
-
-#include <sys/time.h>
-#include <sys/types.h>
-#include <sys/unistd.h>
-
-void usleep(unsigned long time)
+// The provided string is consumed!
+std::string wordwrap_line(std::string &s, int width, bool tags)
 {
-    struct timeval timer;
+    const char *cp0 = s.c_str();
+    const char *cp = cp0, *space = 0;
+    ucs_t c;
 
-    timer.tv_sec  = (time / 1000000L);
-    timer.tv_usec = (time % 1000000L);
+    while (int clen = utf8towc(&c, cp))
+    {
+        int cw = wcwidth(c);
+        if (c == ' ')
+            space = cp;
+        else if (c == '\n')
+        {
+            space = cp;
+            break;
+        }
+        if (c == '<' && tags)
+        {
+            ASSERT(cw == 1);
+            if (cp[1] == '<') // "<<" escape
+            {
+                // Note: this must be after a possible wrap, otherwise we could
+                // split the escape between lines.
+                cp++;
+            }
+            else
+            {
+                cw = 0;
+                // Skip the whole tag.
+                while (*cp != '>')
+                {
+                    if (!*cp)
+                    {
+                        // Everything so far fitted, report error.
+                        std::string ret = s + ">";
+                        s = "<lightred>ERROR: string above had unterminated tag</lightred>";
+                        return ret;
+                    }
+                    cp++;
+                }
+            }
+        }
 
-    select(0, NULL, NULL, NULL, &timer);
-}
-#endif
+        if (cw > width)
+            break;
 
-// Not the greatest version of snprintf, but a functional one that's
-// a bit safer than raw sprintf().  Note that this doesn't do the
-// special behaviour for size == 0, largely because the return value
-// in that case varies depending on which standard is being used (SUSv2
-// returns an unspecified value < 1, whereas C99 allows str == NULL
-// and returns the number of characters that would have been written). -- bwr
-#ifdef NEED_SNPRINTF
+        if (cw >= 0)
+            width -= cw;
+        cp += clen;
+    }
 
-#include <string.h>
+    if (!c)
+    {
+        // everything fits
+        std::string ret = s;
+        s.clear();
+        return ret;
+    }
 
-int snprintf( char *str, size_t size, const char *format, ... )
-{
-    va_list argp;
-    va_start( argp, format );
+    if (space)
+        cp = space;
+    const std::string ret = s.substr(0, cp - cp0);
 
-    char *buff = new char [ 10 * size ];  // hopefully enough
-    if (!buff)
-        end(1, false, "Out of memory\n");
+    // eat all trailing spaces and up to one newline
+    while (*cp == ' ')
+        cp++;
+    if (*cp == '\n')
+        cp++;
+    s.erase(0, cp - cp0);
 
-    vsprintf( buff, format, argp );
-    strncpy( str, buff, size );
-    str[ size - 1 ] = 0;
-
-    int ret = strlen( str );
-    if ((unsigned int) ret == size - 1 && strlen( buff ) >= size)
-        ret = -1;
-
-    delete [] buff;
-
-    va_end( argp );
-
-    return (ret);
+    return ret;
 }
 
-#endif
+/**
+ * Compare two strings, sorting integer numeric parts according to their value.
+ *
+ * "foo123bar" > "foo99bar"
+ * "0.10" > "0.9" (version sort)
+ *
+ * @param limit If passed, comparison ends after X numeric parts.
+ * @return As in strcmp().
+**/
+int numcmp(const char *a, const char *b, int limit)
+{
+    int res;
+
+not_numeric:
+    while (*a && *a == *b && !isadigit(*a))
+    {
+        a++;
+        b++;
+    }
+    if (!a && !b)
+        return 0;
+    if (!isadigit(*a) || !isadigit(*b))
+        return (*a < *b) ? -1 : (*a > *b) ? 1 : 0;
+    while (*a == '0')
+        a++;
+    while (*b == '0')
+        b++;
+    res = 0;
+    while (isadigit(*a))
+    {
+        if (!isadigit(*b))
+            return 1;
+        if (*a != *b && !res)
+            res = (*a < *b) ? -1 : 1;
+        a++;
+        b++;
+    }
+    if (isadigit(*b))
+        return -1;
+    if (res)
+        return res;
+
+    if (--limit)
+        goto not_numeric;
+    return 0;
+}
+
+// make STL sort happy
+bool numcmpstr(const std::string a, const std::string b)
+{
+    return numcmp(a.c_str(), b.c_str()) == -1;
+}
+
+bool version_is_stable(const char *v)
+{
+    // vulnerable to changes in the versioning scheme
+    for (;; v++)
+    {
+        if (*v == '.' || isadigit(*v))
+            continue;
+        if (*v == '-')
+            return isadigit(v[1]);
+        return true;
+    }
+}
 
 #ifndef USE_TILE
-void cgotoxy(int x, int y, int region)
+coord_def cgettopleft(GotoRegion region)
 {
-    ASSERT(x >= 1);
-    ASSERT(y >= 1);
     switch (region)
     {
     case GOTO_MLIST:
-        ASSERT(x <= crawl_view.mlistsz.x);
-        ASSERT(y <= crawl_view.mlistsz.y);
-        gotoxy_sys(x + crawl_view.mlistp.x - 1, y + crawl_view.mlistp.y - 1);
-        break;
+        return crawl_view.mlistp;
     case GOTO_STAT:
-        ASSERT(x <= crawl_view.hudsz.x);
-        ASSERT(y <= crawl_view.hudsz.y);
-        gotoxy_sys(x + crawl_view.hudp.x - 1, y + crawl_view.hudp.y - 1);
-        break;
+        return crawl_view.hudp;
     case GOTO_MSG:
-        ASSERT(x <= crawl_view.msgsz.x);
-        ASSERT(y <= crawl_view.msgsz.y);
-        gotoxy_sys(x + crawl_view.msgp.x - 1, y + crawl_view.msgp.y - 1);
-        break;
+        return crawl_view.msgp;
+    case GOTO_DNGN:
+        return crawl_view.viewp;
     case GOTO_CRT:
     default:
-        gotoxy_sys(x, y);
-        break;
+        return crawl_view.termp;
     }
 }
-#endif
-///////////////////////////////////////////////////////////////////////
-// Pattern matching
 
-inline int pm_lower(int ch, bool icase)
+coord_def cgetpos(GotoRegion region)
 {
-    return icase? tolower(ch) : ch;
+    const coord_def where = coord_def(wherex(), wherey());
+    return (where - cgettopleft(region) + coord_def(1, 1));
 }
 
-// Determines whether the pattern specified by 'pattern' matches the given
-// text. A pattern is a simple glob, with the traditional * and ? wildcards.
-static bool glob_match( const char *pattern, const char *text, bool icase )
-{
-    char p, t;
-    bool special;
+static GotoRegion _current_region = GOTO_CRT;
 
-    while (true)
+void cgotoxy(int x, int y, GotoRegion region)
+{
+    _current_region = region;
+    const coord_def tl = cgettopleft(region);
+    const coord_def sz = cgetsize(region);
+
+    ASSERT_SAVE(x >= 1 && x <= sz.x);
+    ASSERT_SAVE(y >= 1 && y <= sz.y);
+
+    gotoxy_sys(tl.x + x - 1, tl.y + y - 1);
+}
+
+GotoRegion get_cursor_region()
+{
+    return (_current_region);
+}
+#endif // USE_TILE
+
+coord_def cgetsize(GotoRegion region)
+{
+    switch (region)
     {
-        p = pm_lower(*pattern++, icase);
-        t = pm_lower(*text++, icase);
-        special = true;
-
-        if (!p)
-            return (t == 0);
-
-        if (p == '\\' && *pattern)
-        {
-            p       = pm_lower(*pattern++, icase);
-            special = false;
-        }
-
-        if (p == '*' && special)
-        {
-            // Try to match exactly at the current text position...
-            if (!*pattern || glob_match(pattern, text - 1, icase))
-                return (true);
-
-            // Or skip one character in the text and try the wildcard match
-            // again. If this is the end of the text, the match has failed.
-            return (t ? glob_match(pattern - 1, text, icase) : false);
-        }
-        else if (!t || p != t && (p != '?' || !special))
-            return (false);
+    case GOTO_MLIST:
+        return crawl_view.mlistsz;
+    case GOTO_STAT:
+        return crawl_view.hudsz;
+    case GOTO_MSG:
+        return crawl_view.msgsz;
+    case GOTO_DNGN:
+        return crawl_view.viewsz;
+    case GOTO_CRT:
+    default:
+        return crawl_view.termsz;
     }
 }
 
-////////////////////////////////////////////////////////////////////
-// Basic glob (always available)
-
-struct glob_info
+void cscroll(int n, GotoRegion region)
 {
-    std::string s;
-    bool ignore_case;
-};
-
-void *compile_glob_pattern(const char *pattern, bool icase)
-{
-    // If we're using simple globs, we need to box the pattern with '*'
-    std::string s = std::string("*") + pattern + "*";
-    glob_info *gi = new glob_info;
-    if (gi)
-    {
-        gi->s = s;
-        gi->ignore_case = icase;
-    }
-    return gi;
+    // only implemented for the message window right now
+    if (region == GOTO_MSG)
+        scroll_message_window(n);
 }
 
-void free_compiled_glob_pattern(void *compiled_pattern)
-{
-    delete static_cast<glob_info *>( compiled_pattern );
-}
-
-bool glob_pattern_match(void *compiled_pattern, const char *text, int length)
-{
-    glob_info *gi = static_cast<glob_info *>( compiled_pattern );
-    return glob_match(gi->s.c_str(), text, gi->ignore_case);
-}
-////////////////////////////////////////////////////////////////////
-
-#if defined(REGEX_PCRE)
-////////////////////////////////////////////////////////////////////
-// Perl Compatible Regular Expressions
-
-void *compile_pattern(const char *pattern, bool icase)
-{
-    const char *error;
-    int erroffset;
-    int flags = icase? PCRE_CASELESS : 0;
-    return pcre_compile(pattern,
-                        flags,
-                        &error,
-                        &erroffset,
-                        NULL);
-}
-
-void free_compiled_pattern(void *cp)
-{
-    if (cp)
-        pcre_free(cp);
-}
-
-bool pattern_match(void *compiled_pattern, const char *text, int length)
-{
-    int ovector[42];
-    int pcre_rc = pcre_exec(static_cast<pcre *>(compiled_pattern),
-                            NULL,
-                            text, length, 0, 0,
-                            ovector, sizeof(ovector) / sizeof(*ovector));
-    return (pcre_rc >= 0);
-}
-
-////////////////////////////////////////////////////////////////////
-#elif defined(REGEX_POSIX)
-////////////////////////////////////////////////////////////////////
-// POSIX regular expressions
-
-void *compile_pattern(const char *pattern, bool icase)
-{
-    regex_t *re = new regex_t;
-    if (!re)
-        return NULL;
-
-    int flags = REG_EXTENDED | REG_NOSUB;
-    if (icase)
-        flags |= REG_ICASE;
-    int rc = regcomp(re, pattern, flags);
-    // Nonzero return code == failure
-    if (rc)
-    {
-        delete re;
-        return NULL;
-    }
-    return re;
-}
-
-void free_compiled_pattern(void *cp)
-{
-    if (cp)
-    {
-        regex_t *re = static_cast<regex_t *>( cp );
-        regfree(re);
-        delete re;
-    }
-}
-
-bool pattern_match(void *compiled_pattern, const char *text, int length)
-{
-    regex_t *re = static_cast<regex_t *>( compiled_pattern );
-    return !regexec(re, text, 0, NULL, 0);
-}
-
-////////////////////////////////////////////////////////////////////
-#else
-
-void *compile_pattern(const char *pattern, bool icase)
-{
-    return compile_glob_pattern(pattern, icase);
-}
-
-void free_compiled_pattern(void *cp)
-{
-    free_compiled_glob_pattern(cp);
-}
-
-bool pattern_match(void *compiled_pattern, const char *text, int length)
-{
-    return glob_pattern_match(compiled_pattern, text, length);
-}
-
-#endif
 
 mouse_mode mouse_control::ms_current_mode = MOUSE_MODE_NORMAL;
+
+size_t strlcpy(char *dst, const char *src, size_t n)
+{
+    if (!n)
+        return strlen(src);
+
+    const char *s = src;
+
+    while (--n > 0)
+        if (!(*dst++ = *s++))
+            break;
+
+    if (!n)
+    {
+        *dst++ = 0;
+        while (*s++)
+            ;
+    }
+
+    return s - src - 1;
+}
+
+#ifdef TARGET_OS_WINDOWS
+// FIXME: This function should detect if aero is running, but the DwmIsCompositionEnabled
+// function isn't included in msys, so I don't know how to do that. Instead, I just check
+// if we are running vista or higher. -rla
+bool _is_aero()
+{
+    OSVERSIONINFOEX osvi;
+    osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);
+    if (GetVersionEx((OSVERSIONINFO *) &osvi))
+        return (osvi.dwMajorVersion >= 6);
+    else
+        return false;
+}
+
+taskbar_pos get_taskbar_pos()
+{
+    RECT rect;
+    HWND taskbar = FindWindow("Shell_traywnd", NULL);
+    if (taskbar && GetWindowRect(taskbar, &rect))
+    {
+        if (rect.right - rect.left > rect.bottom - rect.top)
+        {
+            if (rect.top > 0)
+                return TASKBAR_BOTTOM;
+            else
+                return TASKBAR_TOP;
+        }
+        else
+        {
+            if (rect.left > 0)
+                return TASKBAR_RIGHT;
+            else
+                return TASKBAR_LEFT;
+        }
+    }
+    return TASKBAR_NO;
+}
+
+int get_taskbar_size()
+{
+    RECT rect;
+    int size;
+    taskbar_pos tpos = get_taskbar_pos();
+    HWND taskbar = FindWindow("Shell_traywnd", NULL);
+
+    if (taskbar && GetWindowRect(taskbar, &rect))
+    {
+        if (tpos & TASKBAR_H)
+                size = rect.bottom - rect.top;
+        else if (tpos & TASKBAR_V)
+                size = rect.right - rect.left;
+        else
+            return 0;
+
+        if (_is_aero())
+            size += 3; // Taskbar behave strangely when aero is active.
+
+        return size;
+    }
+    return 0;
+}
+
+static BOOL WINAPI console_handler(DWORD sig)
+{
+    switch(sig)
+    {
+    case CTRL_C_EVENT:
+    case CTRL_BREAK_EVENT:
+        return true; // block the signal
+    default:
+        return false;
+    case CTRL_CLOSE_EVENT:
+    case CTRL_LOGOFF_EVENT:
+    case CTRL_SHUTDOWN_EVENT:
+        if (crawl_state.seen_hups++)
+            return true;
+
+        sighup_save_and_exit();
+        return true;
+    }
+}
+
+void init_signals()
+{
+    // If there's no console,
+    SetConsoleCtrlHandler(console_handler, true);
+}
+
+#else
+
+/* [ds] This SIGHUP handling is primitive and far from safe, but it
+ * should be better than nothing. Feel free to get rigorous on this.
+ */
+static void handle_hangup(int)
+{
+    if (crawl_state.seen_hups++)
+        return;
+
+#ifdef USE_TILE
+    // XXX: Will a tiles build ever need to handle the HUP signal?
+    sighup_save_and_exit();
+#elif defined(USE_CURSES)
+    // When using Curses, closing stdin will cause any Curses call blocking
+    // on key-presses to immediately return, including any call that was
+    // still blocking in the main thread when the HUP signal was caught.
+    // This should guarantee that the main thread will un-stall and
+    // will eventually return to _input() in main.cc, which will then
+    // call sighup_save_and_exit().
+    //
+    // The point to all this is that if a user is playing a game on a
+    // remote server and disconnects at a --more-- prompt, that when
+    // the player reconnects the code behind the more() call will execute
+    // before the disconnected game is saved, thus (for example) preventing
+    // the hack of avoiding excomunication consesquences because of the
+    // more() after "You have lost your religion!"
+    fclose(stdin);
+#else
+     #error "Must use either Curses or tiles on Unix"
+#endif
+}
+
+void init_signals()
+{
+#ifdef DGAMELAUNCH
+    // Force timezone to UTC.
+    setenv("TZ", "", 1);
+    tzset();
+#endif
+
+#ifdef USE_UNIX_SIGNALS
+#ifdef SIGQUIT
+    signal(SIGQUIT, SIG_IGN);
+#endif
+
+#ifdef SIGINT
+    signal(SIGINT, SIG_IGN);
+#endif
+
+    signal(SIGHUP, handle_hangup);
+#endif
+
+#ifdef DGL_ENABLE_CORE_DUMP
+    rlimit lim;
+    if (!getrlimit(RLIMIT_CORE, &lim))
+    {
+        lim.rlim_cur = RLIM_INFINITY;
+        setrlimit(RLIMIT_CORE, &lim);
+    }
+#endif
+}
+
+#endif
