@@ -1,13 +1,9 @@
-/*
- *  File:       spl-util.cc                                          *
- *  Summary:    data handlers for player-avilable spell list         *
- *  Written by: don brodale <dbrodale@bigfootinteractive.com>        *
- *                                                                   *
- *  Modified for Crawl Reference by $Author$ on $Date$
- */
+/**
+ * @file
+ * @brief data handlers for player-available spell list
+**/
 
 #include "AppHdr.h"
-REVISION("$Rev$");
 
 #include "spl-util.h"
 
@@ -22,27 +18,29 @@ REVISION("$Rev$");
 #include "externs.h"
 
 #include "beam.h"
+#include "coord.h"
+#include "coordit.h"
 #include "directn.h"
 #include "debug.h"
+#include "godabil.h"
 #include "stuff.h"
-#include "itemname.h"
-#include "macro.h"
-#include "misc.h"
-#include "monstuff.h"
+#include "env.h"
+#include "libutil.h"
+#include "mon-behv.h"
+#include "mon-util.h"
 #include "notes.h"
+#include "options.h"
 #include "player.h"
-#include "spells4.h"
-#include "spl-book.h"
+#include "religion.h"
 #include "spl-cast.h"
-#include "spl-util.h"
+#include "spl-book.h"
+#include "spl-damage.h"
+#include "spl-zap.h"
 #include "terrain.h"
-#include "view.h"
+#include "item_use.h"
+#include "transform.h"
 
 
-#ifdef DOS
-#include <conio.h>
-#endif
-
 struct spell_desc
 {
     int id;
@@ -50,13 +48,17 @@ struct spell_desc
     unsigned int disciplines; // bitfield
     unsigned int flags;       // bitfield
     unsigned int level;
+
+    // Usually in the range 0..200 (0 means uncapped).
+    // Note that some spells are also capped through zap_type.
+    // See spell_power_cap below.
     int power_cap;
 
     // At power 0, you get min_range. At power power_cap, you get max_range.
     int min_range;
     int max_range;
 
-    // How much louder or quieter the spell is than the default.
+    // Modify spell level for spell noise purposes.
     int noise_mod;
 
     const char  *target_prompt;
@@ -68,8 +70,9 @@ struct spell_desc
     bool         ms_utility;
 };
 
-static struct spell_desc spelldata[] = {
-#include "spl-data.h"
+static struct spell_desc spelldata[] =
+{
+    #include "spl-data.h"
 };
 
 static int spell_list[NUM_SPELLS];
@@ -77,10 +80,6 @@ static int spell_list[NUM_SPELLS];
 #define SPELLDATASIZE (sizeof(spelldata)/sizeof(struct spell_desc))
 
 static struct spell_desc *_seekspell(spell_type spellid);
-static bool _cloud_helper(cloud_func func, const coord_def& where,
-                          int pow, int spread_rate,
-                          cloud_type ctype, kill_category whose,
-                          killer_type killer);
 
 //
 //             BEGIN PUBLIC FUNCTIONS
@@ -96,11 +95,11 @@ void init_spell_descs(void)
     {
         spell_desc &data = spelldata[i];
 
-#if DEBUG
+#ifdef DEBUG
         if (data.id < SPELL_NO_SPELL || data.id >= NUM_SPELLS)
             end(1, false, "spell #%d has invalid id %d", i, data.id);
 
-        if (data.title == NULL || strlen(data.title) == 0)
+        if (data.title == NULL || !*data.title)
             end(1, false, "spell #%d, id %d has no name", i, data.id);
 
         if (data.level < 1 || data.level > 9)
@@ -165,7 +164,7 @@ spell_type spell_by_name(std::string name, bool partial_match)
         return (SPELL_NO_SPELL);
     }
 
-    int spellmatch = -1;
+    spell_type spellmatch = SPELL_NO_SPELL;
     for (int i = 0; i < NUM_SPELLS; i++)
     {
         spell_type type = static_cast<spell_type>(i);
@@ -178,14 +177,13 @@ spell_type spell_by_name(std::string name, bool partial_match)
         if (spell_name.find(name) != std::string::npos)
         {
             if (spell_name == name)
-                return static_cast<spell_type>(i);
+                return type;
 
-            spellmatch = i;
+            spellmatch = type;
         }
     }
 
-    return (spellmatch != -1 ? static_cast<spell_type>(spellmatch)
-                             : SPELL_NO_SPELL);
+    return spellmatch;
 }
 
 spschool_flag_type school_by_name(std::string name)
@@ -239,11 +237,11 @@ spschool_flag_type school_by_name(std::string name)
    return SPTYP_NONE;
 }
 
-int get_spell_slot_by_letter( char letter )
+int get_spell_slot_by_letter(char letter)
 {
-    ASSERT( isalpha( letter ) );
+    ASSERT(isaalpha(letter));
 
-    const int index = letter_to_index( letter );
+    const int index = letter_to_index(letter);
 
     if (you.spell_letter_table[ index ] == -1)
         return (-1);
@@ -251,21 +249,37 @@ int get_spell_slot_by_letter( char letter )
     return (you.spell_letter_table[index]);
 }
 
-spell_type get_spell_by_letter( char letter )
+int get_spell_slot(spell_type spell)
 {
-    ASSERT( isalpha( letter ) );
+    for (int i = 0; i < MAX_KNOWN_SPELLS; i++)
+        if (you.spells[i] == spell)
+            return i;
 
-    const int slot = get_spell_slot_by_letter( letter );
+    return -1;
+}
+
+int get_spell_letter(spell_type spell)
+{
+    return index_to_letter(get_spell_slot(spell));
+}
+
+spell_type get_spell_by_letter(char letter)
+{
+    ASSERT(isaalpha(letter));
+
+    const int slot = get_spell_slot_by_letter(letter);
 
     return ((slot == -1) ? SPELL_NO_SPELL : you.spells[slot]);
 }
 
-bool add_spell_to_memory( spell_type spell )
+bool add_spell_to_memory(spell_type spell)
 {
-    int i, j;
-
+    int i;
+    int j = -1;
+    std::string sname = spell_title(spell);
+    lowercase(sname);
     // first we find a slot in our head:
-    for (i = 0; i < 25; i++)
+    for (i = 0; i < MAX_KNOWN_SPELLS; i++)
     {
         if (you.spells[i] == SPELL_NO_SPELL)
             break;
@@ -274,11 +288,30 @@ bool add_spell_to_memory( spell_type spell )
     you.spells[i] = spell;
 
     // now we find an available label:
-    for (j = 0; j < 52; j++)
+    // first check to see whether we've chosen an automatic label:
+    for (unsigned k = 0; k < Options.auto_spell_letters.size(); ++k)
     {
-        if (you.spell_letter_table[j] == -1)
+        if (!Options.auto_spell_letters[k].first.matches(sname))
+            continue;
+        for (unsigned l = 0; l < Options.auto_spell_letters[k].second.length(); ++l)
+            if (isaalpha(Options.auto_spell_letters[k].second[l]) &&
+                you.spell_letter_table[letter_to_index(Options.auto_spell_letters[k].second[l])] == -1)
+            {
+                j = letter_to_index(Options.auto_spell_letters[k].second[l]);
+                mprf("Spell assigned to '%c'.",
+                     Options.auto_spell_letters[k].second[l]);
+                break;
+            }
+        if (j != -1)
             break;
     }
+    // If we didn't find a label above, choose the first available one.
+    if (j == -1)
+        for (j = 0; j < 52; j++)
+        {
+            if (you.spell_letter_table[j] == -1)
+                break;
+        }
 
     you.spell_letter_table[j] = i;
 
@@ -286,13 +319,23 @@ bool add_spell_to_memory( spell_type spell )
 
     take_note(Note(NOTE_LEARN_SPELL, spell));
 
+#ifdef USE_TILE_LOCAL
+    tiles.layout_statcol();
+    redraw_screen();
+#endif
+
     return (true);
 }
 
-bool del_spell_from_memory_by_slot( int slot )
+bool del_spell_from_memory_by_slot(int slot)
 {
+    ASSERT(slot >= 0 && slot < MAX_KNOWN_SPELLS);
     int j;
 
+    if (you.last_cast_spell == you.spells[slot])
+        you.last_cast_spell = SPELL_NO_SPELL;
+
+    mprf("Your memory of %s unravels.", spell_title(you.spells[slot]));
     you.spells[ slot ] = SPELL_NO_SPELL;
 
     for (j = 0; j < 52; j++)
@@ -304,11 +347,24 @@ bool del_spell_from_memory_by_slot( int slot )
 
     you.spell_no--;
 
+#ifdef USE_TILE_LOCAL
+    tiles.layout_statcol();
+    redraw_screen();
+#endif
+
     return (true);
 }
 
+bool del_spell_from_memory(spell_type spell)
+{
+    int i = get_spell_slot(spell);
+    if (i == -1)
+        return (false);
+    else
+        return del_spell_from_memory_by_slot(i);
+}
 
-int spell_hunger(spell_type which_spell)
+int spell_hunger(spell_type which_spell, bool rod)
 {
     const int level = spell_difficulty(which_spell);
 
@@ -323,7 +379,13 @@ int spell_hunger(spell_type which_spell)
     else
         hunger = (basehunger[0] * level * level) / 4;
 
-    hunger -= you.intel * you.skills[SK_SPELLCASTING];
+    if (rod)
+    {
+        hunger -= 10 * you.skill(SK_EVOCATIONS);
+        hunger = std::max(hunger, level * 5);
+    }
+    else
+        hunger -= you.intel() * you.skill(SK_SPELLCASTING);
 
     if (hunger < 0)
         hunger = 0;
@@ -387,6 +449,15 @@ bool spell_sanctuary_castable(spell_type spell)
 // for Xom acting (more power = more likely to grab his attention) {dlb}
 int spell_mana(spell_type which_spell)
 {
+    if (vehumet_supports_spell(which_spell)
+        && you.religion == GOD_VEHUMET
+        && !player_under_penance()
+        && you.piety >= piety_breakpoint(3)
+        && _seekspell(which_spell)->level >= 5)
+    {
+        return (_seekspell(which_spell)->level - 1);
+    }
+
     return (_seekspell(which_spell)->level);
 }
 
@@ -397,17 +468,17 @@ int spell_difficulty(spell_type which_spell)
     return (_seekspell(which_spell)->level);
 }
 
-int spell_levels_required( spell_type which_spell )
+int spell_levels_required(spell_type which_spell)
 {
-    int levels = spell_difficulty( which_spell );
+    int levels = spell_difficulty(which_spell);
 
     if (which_spell == SPELL_DELAYED_FIREBALL
-        && player_has_spell( SPELL_FIREBALL ))
+        && you.has_spell(SPELL_FIREBALL))
     {
-        levels -= spell_difficulty( SPELL_FIREBALL );
+        levels -= spell_difficulty(SPELL_FIREBALL);
     }
     else if (which_spell == SPELL_FIREBALL
-            && player_has_spell( SPELL_DELAYED_FIREBALL ))
+            && you.has_spell(SPELL_DELAYED_FIREBALL))
     {
         levels = 0;
     }
@@ -415,25 +486,29 @@ int spell_levels_required( spell_type which_spell )
     return (levels);
 }
 
-unsigned int get_spell_flags( spell_type which_spell )
+unsigned int get_spell_flags(spell_type which_spell)
 {
     return (_seekspell(which_spell)->flags);
 }
 
-const char *get_spell_target_prompt( spell_type which_spell )
+const char *get_spell_target_prompt(spell_type which_spell)
 {
     return (_seekspell(which_spell)->target_prompt);
 }
 
 bool spell_typematch(spell_type which_spell, unsigned int which_discipline)
 {
-    return (_seekspell(which_spell)->disciplines & which_discipline);
+    return (get_spell_disciplines(which_spell) & which_discipline);
 }
 
 //jmf: next two for simple bit handling
 unsigned int get_spell_disciplines(spell_type spell)
 {
-    return (_seekspell(spell)->disciplines);
+    unsigned int dis = _seekspell(spell)->disciplines;
+    if (spell == SPELL_DRAGON_FORM && player_genus(GENPC_DRACONIAN))
+        dis &= (~SPTYP_FIRE);
+
+    return dis;
 }
 
 int count_bits(unsigned int bits)
@@ -454,7 +529,7 @@ bool disciplines_conflict(unsigned int disc1, unsigned int disc2)
 {
     const unsigned int combined = disc1 | disc2;
 
-    return (   (combined & SPTYP_EARTH) && (combined & SPTYP_AIR)
+    return ((combined & SPTYP_EARTH) && (combined & SPTYP_AIR)
             || (combined & SPTYP_FIRE)  && (combined & SPTYP_ICE)
             || (combined & SPTYP_HOLY)  && (combined & SPTYP_NECROMANCY));
 }
@@ -474,14 +549,22 @@ const char *spell_title(spell_type spell)
 // Apply a function-pointer to all visible squares
 // Returns summation of return values from passed in function.
 int apply_area_visible(cell_func cf, int power,
-                       bool pass_through_trans, actor *agent)
+                       bool pass_through_trans, actor *agent,
+                       bool affect_scryed)
 {
     int rv = 0;
 
+    bool xray = you.xray_vision;
+
+    if (!affect_scryed)
+        you.xray_vision = false;
+
     for (radius_iterator ri(you.pos(), LOS_RADIUS); ri; ++ri)
-        if (pass_through_trans || see_grid_no_trans(*ri))
+        if (pass_through_trans || you.see_cell_no_trans(*ri))
             rv += cf(*ri, power, 0, agent);
 
+    you.xray_vision = xray;
+
     return (rv);
 }
 
@@ -512,6 +595,26 @@ int apply_area_around_square(cell_func cf, const coord_def& where, int power,
     return (rv);
 }
 
+// Like apply_area_around_square, but for monsters in those squares,
+// and takes care not to affect monsters twice that change position.
+int apply_monsters_around_square(monster_func mf, const coord_def& where,
+                                  int power)
+{
+    int rv = 0;
+    std::set<const monster*> affected;
+    for (adjacent_iterator ai(where, true); ai; ++ai)
+    {
+        monster* mon = monster_at(*ai);
+        if (mon && affected.find(mon) == affected.end())
+        {
+            rv += mf(mon, power);
+            affected.insert(mon);
+        }
+    }
+
+   return (rv);
+}
+
 // Affect up to max_targs monsters around a point, chosen randomly.
 // Return varies with the function called; return values will be added up.
 int apply_random_around_square(cell_func cf, const coord_def& where,
@@ -640,9 +743,12 @@ int apply_random_around_square(cell_func cf, const coord_def& where,
 int apply_one_neighbouring_square(cell_func cf, int power, actor *agent)
 {
     dist bmove;
+    direction_chooser_args args;
+    args.restricts = DIR_DIR;
+    args.mode = TARG_ANY;
 
     mpr("Which direction? [ESC to cancel]", MSGCH_PROMPT);
-    direction(bmove, DIR_DIR, TARG_ENEMY);
+    direction(bmove, args);
 
     if (!bmove.isValid)
     {
@@ -650,19 +756,13 @@ int apply_one_neighbouring_square(cell_func cf, int power, actor *agent)
         return (-1);
     }
 
-    int rv = cf(you.pos() + bmove.delta, power, 1, agent);
-
-    if (rv == 0)
-        canned_msg(MSG_NOTHING_HAPPENS);
-
-    return (rv);
+    return cf(you.pos() + bmove.delta, power, 1, agent);
 }
 
 int apply_area_within_radius(cell_func cf, const coord_def& where,
                              int pow, int radius, int ctype,
                              actor *agent)
 {
-
     int rv = 0;
 
     for (radius_iterator ri(where, radius, false, false); ri; ++ri)
@@ -671,89 +771,69 @@ int apply_area_within_radius(cell_func cf, const coord_def& where,
     return (rv);
 }
 
-// apply_area_cloud:
-// Try to make a realistic cloud by expanding from a point, filling empty
-// floor tiles until we run out of material (passed in as number).
-// We really need some sort of a queue structure, since ideally I'd like
-// to do a (shallow) breadth-first-search of the dungeon floor.
-// This ought to work okay for small clouds.
-void apply_area_cloud( cloud_func func, const coord_def& where,
+void apply_area_cloud(cloud_func func, const coord_def& where,
                        int pow, int number, cloud_type ctype,
-                       kill_category whose, killer_type killer,
-                       int spread_rate )
+                       const actor *agent,
+                       int spread_rate, int colour, std::string name,
+                       std::string tile)
 {
-    int good_squares = 0;
-    int neighbours[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
-
-    if (number && _cloud_helper(func, where, pow, spread_rate, ctype, whose,
-                                killer))
-        number--;
-
-    if (number == 0)
+    if (number <= 0)
         return;
 
-    // These indices depend on the order in Compass (see acr.cc)
-    int compass_order_orth[4] = { 2, 6, 4, 0 };
-    int compass_order_diag[4] = { 1, 3, 5, 7 };
-
-    int* const arrs[2] = { compass_order_orth, compass_order_diag };
-
-    for ( int m = 0; m < 2; ++m )
+    targetter_cloud place(0, 0, number, number);
+    if (!place.set_aim(where))
+        return;
+    unsigned int dist = 0;
+    while (number > 0)
     {
-        // Randomise, but do orthogonals first and diagonals later.
-        std::random_shuffle( arrs[m], arrs[m] + 4 );
-        for ( int i = 0; i < 4 && number; ++i )
-        {
-            const int aux = arrs[m][i];
-            if ( _cloud_helper(func, where + Compass[aux],
-                               pow, spread_rate, ctype, whose, killer))
-            {
-                number--;
-                good_squares++;
-                neighbours[aux]++;
-            }
-        }
-    }
+        while (place.queue[dist].empty())
+            if (++dist >= place.queue.size())
+                return;
+        std::vector<coord_def> &q = place.queue[dist];
+        int el = random2(q.size());
+        coord_def c = q[el];
+        q[el] = q[q.size() - 1];
+        q.pop_back();
 
-    // Get a random permutation.
-    int perm[8];
-    for ( int i = 0; i < 8; ++i )
-        perm[i] = i;
-    std::random_shuffle(perm, perm+8);
-    for (int i = 0; i < 8 && number; i++)
-    {
-        // Spread (in random order.)
-        const int j = perm[i];
-
-        if (neighbours[j] == 0)
+        if (place.seen[c] <= 0)
             continue;
-
-        int spread = number / good_squares;
-        number -= spread;
-        good_squares--;
-        apply_area_cloud(func, where + Compass[j], pow, spread, ctype, whose,
-                         killer, spread_rate);
+        func(c, pow, spread_rate, ctype, agent, colour, name, tile);
+        number--;
     }
 }
 
 // Select a spell direction and fill dist and pbolt appropriately.
 // Return false if the user canceled, true otherwise.
-bool spell_direction( dist &spelld, bolt &pbolt,
+// FIXME: this should accept a direction_chooser_args directly rather
+// than move the arguments into one.
+bool spell_direction(dist &spelld, bolt &pbolt,
                       targeting_type restrict, targ_mode_type mode,
                       int range,
                       bool needs_path, bool may_target_monster,
-                      bool may_target_self, const char *prompt,
-                      bool cancel_at_self )
+                      bool may_target_self, const char *target_prefix,
+                      const char* top_prompt, bool cancel_at_self,
+                      targetter *hitfunc, desc_filter get_desc_func)
 {
-    if (restrict != DIR_DIR)
-        message_current_target();
-
     if (range < 1)
         range = (pbolt.range < 1) ? LOS_RADIUS : pbolt.range;
 
-    direction( spelld, restrict, mode, range, false, needs_path,
-               may_target_monster, may_target_self, prompt, NULL,
-               cancel_at_self );
+    direction_chooser_args args;
+    args.restricts = restrict;
+    args.mode = mode;
+    args.range = range;
+    args.just_looking = false;
+    args.needs_path = needs_path;
+    args.may_target_monster = may_target_monster;
+    args.may_target_self = may_target_self;
+    args.target_prefix = target_prefix;
+    if (top_prompt)
+        args.top_prompt = top_prompt;
+    args.behaviour = NULL;
+    args.cancel_at_self = cancel_at_self;
+    args.hitfunc = hitfunc;
+    args.get_desc_func = get_desc_func;
+
+    direction(spelld, args);
 
     if (!spelld.isValid)
     {
@@ -768,14 +848,16 @@ bool spell_direction( dist &spelld, bolt &pbolt,
     return (true);
 }
 
-const char* spelltype_short_name( int which_spelltype )
+const char* spelltype_short_name(int which_spelltype)
 {
     switch (which_spelltype)
     {
     case SPTYP_CONJURATION:
         return ("Conj");
-    case SPTYP_ENCHANTMENT:
-        return ("Ench");
+    case SPTYP_HEXES:
+        return ("Hex");
+    case SPTYP_CHARMS:
+        return ("Chrm");
     case SPTYP_FIRE:
         return ("Fire");
     case SPTYP_ICE:
@@ -805,14 +887,16 @@ const char* spelltype_short_name( int which_spelltype )
     }
 }
 
-const char* spelltype_long_name( int which_spelltype )
+const char* spelltype_long_name(int which_spelltype)
 {
     switch (which_spelltype)
     {
     case SPTYP_CONJURATION:
         return ("Conjuration");
-    case SPTYP_ENCHANTMENT:
-        return ("Enchantment");
+    case SPTYP_HEXES:
+        return ("Hexes");
+    case SPTYP_CHARMS:
+        return ("Charms");
     case SPTYP_FIRE:
         return ("Fire");
     case SPTYP_ICE:
@@ -842,53 +926,18 @@ const char* spelltype_long_name( int which_spelltype )
     }
 }
 
-const char *spelltype_name(unsigned int which_spelltype)
-{
-    switch (which_spelltype)
-    {
-    case SPTYP_CONJURATION:
-        return ("Conjuration");
-    case SPTYP_ENCHANTMENT:
-        return ("Enchantment");
-    case SPTYP_FIRE:
-        return ("Fire");
-    case SPTYP_ICE:
-        return ("Ice");
-    case SPTYP_TRANSMUTATION:
-        return ("Transmutation");
-    case SPTYP_NECROMANCY:
-        return ("Necromancy");
-    case SPTYP_HOLY:
-        return ("Holy");
-    case SPTYP_SUMMONING:
-        return ("Summoning");
-    case SPTYP_DIVINATION:
-        return ("Divination");
-    case SPTYP_TRANSLOCATION:
-        return ("Translocation");
-    case SPTYP_POISON:
-        return ("Poison");
-    case SPTYP_EARTH:
-        return ("Earth");
-    case SPTYP_AIR:
-        return ("Air");
-    default:
-        return "Buggy";
-    }
-}                               // end spelltype_name()
-
-int spell_type2skill(unsigned int spelltype)
+skill_type spell_type2skill(unsigned int spelltype)
 {
     switch (spelltype)
     {
     case SPTYP_CONJURATION:    return (SK_CONJURATIONS);
-    case SPTYP_ENCHANTMENT:    return (SK_ENCHANTMENTS);
+    case SPTYP_HEXES:          return (SK_HEXES);
+    case SPTYP_CHARMS:         return (SK_CHARMS);
     case SPTYP_FIRE:           return (SK_FIRE_MAGIC);
     case SPTYP_ICE:            return (SK_ICE_MAGIC);
     case SPTYP_TRANSMUTATION:  return (SK_TRANSMUTATIONS);
     case SPTYP_NECROMANCY:     return (SK_NECROMANCY);
     case SPTYP_SUMMONING:      return (SK_SUMMONINGS);
-    case SPTYP_DIVINATION:     return (SK_DIVINATIONS);
     case SPTYP_TRANSLOCATION:  return (SK_TRANSLOCATIONS);
     case SPTYP_POISON:         return (SK_POISON_MAGIC);
     case SPTYP_EARTH:          return (SK_EARTH_MAGIC);
@@ -896,40 +945,11 @@ int spell_type2skill(unsigned int spelltype)
 
     default:
     case SPTYP_HOLY:
-#ifdef DEBUG_DIAGNOSTICS
-        mprf(MSGCH_DIAGNOSTICS, "spell_type2skill: called with spelltype %u",
-             spelltype );
-#endif
-        return (-1);
+    case SPTYP_DIVINATION:
+        dprf("spell_type2skill: called with spelltype %u", spelltype);
+        return (SK_NONE);
     }
-}                               // end spell_type2skill()
-
-int spell_skill2type(unsigned int skill)
-{
-    switch (skill)
-    {
-    case SK_CONJURATIONS:   return (SPTYP_CONJURATION);
-    case SK_ENCHANTMENTS:   return (SPTYP_ENCHANTMENT);
-    case SK_FIRE_MAGIC:     return (SPTYP_FIRE);
-    case SK_ICE_MAGIC:      return (SPTYP_ICE);
-    case SK_TRANSMUTATIONS: return (SPTYP_TRANSMUTATION);
-    case SK_NECROMANCY:     return (SPTYP_NECROMANCY);
-    case SK_SUMMONINGS:     return (SPTYP_SUMMONING);
-    case SK_DIVINATIONS:    return (SPTYP_DIVINATION);
-    case SK_TRANSLOCATIONS: return (SPTYP_TRANSLOCATION);
-    case SK_POISON_MAGIC:   return (SPTYP_POISON);
-    case SK_EARTH_MAGIC:    return (SPTYP_EARTH);
-    case SK_AIR_MAGIC:      return (SPTYP_AIR);
-
-    default:
-    case SPTYP_HOLY:
-#ifdef DEBUG_DIAGNOSTICS
-        mprf(MSGCH_DIAGNOSTICS, "spell_skill2type: called with skill %u",
-             skill);
-#endif
-        return (-1);
-    }
-}                               // end spell_type2skill()
+}
 
 /*
  **************************************************
@@ -955,46 +975,65 @@ bool is_valid_spell(spell_type spell)
             && spell_list[spell] != -1);
 }
 
-static bool _cloud_helper(cloud_func func, const coord_def& where,
-                          int pow, int spread_rate,
-                          cloud_type ctype, kill_category whose,
-                          killer_type killer )
+static bool _spell_range_varies(spell_type spell)
 {
-    if (!grid_is_solid(grd(where)) && env.cgrid(where) == EMPTY_CLOUD)
-    {
-        func(where, pow, spread_rate, ctype, whose, killer);
-        return (true);
-    }
+    int minrange = _seekspell(spell)->min_range;
+    int maxrange = _seekspell(spell)->max_range;
 
-    return (false);
+    return (minrange < maxrange);
 }
 
 int spell_power_cap(spell_type spell)
 {
-    return (_seekspell(spell)->power_cap);
+    const int scap = _seekspell(spell)->power_cap;
+    const int zcap = spell_zap_power_cap(spell);
+
+    if (scap == 0)
+        return (zcap);
+    else if (zcap == 0)
+        return (scap);
+    else
+    {
+        // Two separate power caps; pre-zapping spell power
+        // goes into range.
+        if (scap <= zcap || _spell_range_varies(spell))
+            return (scap);
+        else
+            return (zcap);
+    }
 }
 
-// Sandblast range is 1 if not wielding rocks, 2 if you are.
-static int _sandblast_range(int pow, bool real_cast)
+int spell_range(spell_type spell, int pow, bool real_cast, bool player_spell)
 {
-    int res = 1;
-
-    if (wielding_rocks())
-        res = 2;
-
-    return (res);
-}
-
-
-int spell_range(spell_type spell, int pow, bool real_cast)
-{
-    const int minrange = _seekspell(spell)->min_range;
-    const int maxrange = _seekspell(spell)->max_range;
+    int minrange = _seekspell(spell)->min_range;
+    int maxrange = _seekspell(spell)->max_range;
     ASSERT(maxrange >= minrange);
 
-    // Some cases need to be handled specially.
-    if (spell == SPELL_SANDBLAST)
-        return _sandblast_range(pow, real_cast);
+    // spells with no range have maxrange == minrange == -1
+    if (maxrange < 0)
+        return maxrange;
+
+    // Sandblast is a special case.
+    if (spell == SPELL_SANDBLAST && wielding_rocks())
+    {
+        minrange++;
+        maxrange++;
+    }
+
+    if (player_spell
+        && vehumet_supports_spell(spell)
+        && you.religion == GOD_VEHUMET
+        && spell != SPELL_STICKY_FLAME
+        && spell != SPELL_FREEZE
+        && !player_under_penance()
+        && you.piety >= piety_breakpoint(2))
+    {
+        if (maxrange < LOS_RADIUS)
+            maxrange++;
+
+        if (minrange < LOS_RADIUS)
+            minrange++;
+    }
 
     if (minrange == maxrange)
         return minrange;
@@ -1002,19 +1041,44 @@ int spell_range(spell_type spell, int pow, bool real_cast)
     const int powercap = spell_power_cap(spell);
 
     if (powercap <= pow)
-        return maxrange;
+        return std::min(maxrange, LOS_RADIUS);
 
     // Round appropriately.
-    return ((pow*(maxrange - minrange) + powercap/2) / powercap + minrange);
+    return std::min(LOS_RADIUS,
+           (pow * (maxrange - minrange) + powercap / 2) / powercap + minrange);
 }
 
+/**
+ * Spell casting noise.
+ *
+ * Returns the noise generated by the casting of a spell. The noise depends on
+ * the spell schools and level. A modifier (noise_mod) can be applied to the
+ * spell level.
+ * @see spl-data.h
+ *
+ * \param spell  The spell being casted.
+ * \return       The amount of noise generated.
+**/
 int spell_noise(spell_type spell)
 {
     const spell_desc *desc = _seekspell(spell);
 
-    return desc->noise_mod + spell_noise(desc->disciplines, desc->level);
+    return spell_noise(desc->disciplines, desc->level + desc->noise_mod);
 }
 
+/**
+ * Spell default noise.
+ *
+ * Default value for spell noise given a level and a set of schools.
+ * Formula (use first match):
+ * - Conjuration (noisy)    = \f$ level \f$
+ * - Air and poison (quiet) = \f$ \frac{level}{2} \f$
+ * - Other (normal)         = \f$ \frac{3 \times level}{4} \f$
+ *
+ * \param disciplines  An integer which contain the school flags.
+ * \param level        The level of the spell.
+ * \return             The amount of noise generated.
+**/
 int spell_noise(unsigned int disciplines, int level)
 {
     if (disciplines == SPTYP_NONE)
@@ -1026,3 +1090,244 @@ int spell_noise(unsigned int disciplines, int level)
     else
         return div_round_up(level, 2);
 }
+
+spell_type zap_type_to_spell(zap_type zap)
+{
+    switch (zap)
+    {
+    case ZAP_FLAME:
+        return SPELL_THROW_FLAME;
+    case ZAP_FROST:
+        return SPELL_THROW_FROST;
+    case ZAP_SLOWING:
+        return SPELL_SLOW;
+    case ZAP_HASTING:
+        return SPELL_HASTE;
+    case ZAP_MAGIC_DARTS:
+        return SPELL_MAGIC_DART;
+    case ZAP_HEALING:
+        return SPELL_MAJOR_HEALING;
+    case ZAP_PARALYSIS:
+        return SPELL_PARALYSE;
+    case ZAP_FIRE:
+        return SPELL_BOLT_OF_FIRE;
+    case ZAP_COLD:
+        return SPELL_BOLT_OF_COLD;
+    case ZAP_PRIMAL_WAVE:
+        return SPELL_PRIMAL_WAVE;
+    case ZAP_CONFUSION:
+        return SPELL_CONFUSE;
+    case ZAP_INVISIBILITY:
+        return SPELL_INVISIBILITY;
+    case ZAP_DIGGING:
+        return SPELL_DIG;
+    case ZAP_FIREBALL:
+        return SPELL_FIREBALL;
+    case ZAP_TELEPORTATION:
+        return SPELL_TELEPORT_OTHER;
+    case ZAP_LIGHTNING:
+        return SPELL_LIGHTNING_BOLT;
+    case ZAP_POLYMORPH_OTHER:
+        return SPELL_POLYMORPH_OTHER;
+    case ZAP_NEGATIVE_ENERGY:
+        return SPELL_BOLT_OF_DRAINING;
+    case ZAP_ENSLAVEMENT:
+        return SPELL_ENSLAVEMENT;
+    case ZAP_DISINTEGRATION:
+        return SPELL_DISINTEGRATE;
+    default:
+        die("zap_type_to_spell() only handles wand zaps for now");
+    }
+    return SPELL_NO_SPELL;
+}
+
+bool spell_is_empowered(spell_type spell)
+{
+    if ((you.religion == GOD_VEHUMET)
+        && vehumet_supports_spell(spell)
+        && piety_rank() > 2)
+    {
+        return (true);
+    }
+
+    switch (spell)
+    {
+    case SPELL_SWIFTNESS:
+        // looking at player_movement_speed, this should be correct ~DMB
+        if (player_movement_speed() > 6
+            && you.duration[DUR_CONTROLLED_FLIGHT] > 0
+            && you.duration[DUR_SWIFTNESS] < 1)
+        {
+            return (true);
+        }
+        break;
+    case SPELL_STONESKIN:
+        if (you.duration[DUR_TRANSFORMATION] > 0
+            && you.form == TRAN_STATUE
+            && you.duration[DUR_STONESKIN] < 1)
+        {
+            return (true);
+        }
+        break;
+    case SPELL_OZOCUBUS_ARMOUR:
+        if (you.duration[DUR_TRANSFORMATION] > 0
+            && you.form == TRAN_ICE_BEAST
+            && you.duration[DUR_ICY_ARMOUR] < 1)
+        {
+            return (true);
+        }
+        break;
+    default:
+        break;
+    }
+
+    return (false);
+}
+
+// This function attempts to determine if 'spell' is useless to
+// the player. if 'transient' is true, then it will include checks
+// for volatile or temporary states (such as status effects, mana, etc.)
+//
+// its notably used by 'spell_highlight_by_utility'
+bool spell_is_useless(spell_type spell, bool transient)
+{
+    if (you_cannot_memorise(spell))
+        return (true);
+
+    if (transient)
+    {
+        if (you.duration[DUR_CONF] > 0
+            || spell_mana(spell) > you.magic_points
+            || spell_no_hostile_in_range(spell, get_dist_to_nearest_monster()))
+        {
+            return (true);
+        }
+    }
+
+    switch (spell)
+    {
+    case SPELL_BLINK:
+    case SPELL_CONTROLLED_BLINK:
+    case SPELL_TELEPORT_SELF:
+        if (item_blocks_teleport(false, false))
+            return true;
+        break;
+    case SPELL_SWIFTNESS:
+        // looking at player_movement_speed, this should be correct ~DMB
+        if (player_movement_speed() <= 6)
+            return (true);
+        break;
+    case SPELL_LEVITATION:
+    case SPELL_FLY:
+        if (you.species == SP_KENKU && you.experience_level >= 15)
+            return (true);
+        if (transient && you.is_levitating())
+            return (true);
+        break;
+    case SPELL_INVISIBILITY:
+        if (transient && (you.duration[DUR_INVIS] > 0 || you.backlit()))
+            return (true);
+        break;
+    case SPELL_CONTROL_TELEPORT:
+        if (transient && you.duration[DUR_CONTROL_TELEPORT] > 0)
+            return (true);
+        break;
+    case SPELL_SEE_INVISIBLE:
+        if (you.can_see_invisible(false, false))
+            return (true);
+        break;
+    case SPELL_DARKNESS:
+        // mere corona is not enough, but divine light blocks it completely
+        if (transient && you.haloed())
+            return true;
+        if (you.religion == GOD_SHINING_ONE && !player_under_penance())
+            return true;
+        break;
+    default:
+        break; // quash unhandled constants warnings
+    }
+
+    return (false);
+}
+
+// This function takes a spell, and determines what color it should be
+// highlighted with. You shouldn't have to touch this unless you want
+// to add new highlighting options.
+//
+// as you can see, the functions it uses to determine highlights are:
+//       god_hates_spell(spell, god)
+//       god_likes_spell(spell, god)
+//       spell_is_empowered(spell)
+//       spell_is_useless(spell, transient)
+int spell_highlight_by_utility(spell_type spell, int default_color,
+                               bool transient, bool rod_spell)
+{
+    // If your god hates the spell, that
+    // overrides all other concerns
+    if (god_hates_spell(spell, you.religion))
+        return (COL_FORBIDDEN);
+
+    if (spell_is_empowered(spell) && !rod_spell)
+        default_color = COL_EMPOWERED;
+
+    if (spell_is_useless(spell, transient))
+        default_color = COL_USELESS;
+
+    return (default_color);
+}
+
+bool spell_no_hostile_in_range(spell_type spell, int minRange)
+{
+    if (minRange < 0)
+        return (false);
+
+    bool bonus = 0;
+    switch (spell)
+    {
+    // These don't target monsters.
+    case SPELL_APPORTATION:
+    case SPELL_PROJECTED_NOISE:
+    case SPELL_CONJURE_FLAME:
+    case SPELL_DIG:
+    case SPELL_PASSWALL:
+    case SPELL_GOLUBRIAS_PASSAGE:
+
+    // Airstrike has LOS_RANGE and can go through glass walls.
+    case SPELL_AIRSTRIKE:
+
+    // These bounce and may be aimed elsewhere to bounce at monsters
+    // outside range (I guess).
+    case SPELL_SHOCK:
+    case SPELL_LIGHTNING_BOLT:
+
+    case SPELL_FIRE_STORM:
+        return (false);
+
+    case SPELL_EVAPORATE:
+    case SPELL_MEPHITIC_CLOUD:
+    case SPELL_FIREBALL:
+    case SPELL_FREEZING_CLOUD:
+    case SPELL_NOXIOUS_CLOUD:
+    case SPELL_POISONOUS_CLOUD:
+        // Increase range by one due to cloud radius.
+        bonus = 1;
+        break;
+
+    default:
+        break;
+    }
+
+    // The healing spells.
+    if (testbits(get_spell_flags(spell), SPFLAG_HELPFUL))
+        return (false);
+
+    const int range = calc_spell_range(spell);
+    if (range < 0)
+        return (false);
+
+    const int rsq = (range + bonus) * (range + bonus) + 1;
+    if (rsq < minRange)
+        return (true);
+
+    return (false);
+}
